/* 
 * Copyright (c) 2001 The University of Wroclaw.
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Id: stdlib.c,v 1.5 2001/12/03 09:45:55 malekith Exp $
 *
 * Runtime enviroment for gont.
 *
 * Assumes compiler that allows $ in identifiers, such as GCC.
 * (It is required to have Ksi, so this ain't so important)
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "runtime.h"

void *xmalloc$(ptroff s)
{
	void *p = malloc(s);

	if (p == NULL) {
		fprintf(stderr, "Run out of memory, trying to "
				"allocate %d bytes.\n", (int)s);
		abort();
	} else
		return p;
}

void *make_closure$(ptroff s)
{
	return memset(xmalloc$(s), '\0', s);
}

struct lambda_exp$ *make_lambda$(void *fnc, void *closure)
{
	struct lambda_exp$ *p;
	
	p = xmalloc$(sizeof(*p));
	
	p->fnc = fnc;
	p->closure = closure;
	
	return p;
}

/* Exceptions */

/* This could be highly optimized by using continous chunk of memory for
 * try blocks, they are always pushed and poped in the right order. 
 * However we leave it for now. 
 */

struct try_block$ {
	void *longjmpbuf[5];
	void *value;
	struct try_block$ *chain;
};

static struct try_block$ *try_chain$;

struct try_block$ *push_try_block$(void)
{
	struct try_block$ *t;

	t = xmalloc$(sizeof(*t));
	
	t->value = NULL;
	t->chain = try_chain$;
	
	return (try_chain$ = t);
}

void *pop_try_block$(void)
{
	void *v = try_chain$->value;
	void *f = try_chain$;

	try_chain$ = try_chain$->chain;
	free(f);

	return v;
}

void raise$(void *v)
{
	try_chain$->value = v;
	__builtin_longjmp(try_chain$, 1);
	for (;;) {}
}

/*
 * try B1 with e B2 ->
 *   if (setjmp(push_try_block())) {
 *      e = pop_try_block();
 *      switch e B2, default raise(e)
 *   } else {
 *     B1;
 *     pop_try_block();
 *   }
 * 
 * try B1 with e B2 finally B3 ->
 *   block {
 *     if (setjmp(push_try_block()))
 *       exit(block);
 *     if (setjmp(push_try_block())) {
 *        e = pop_try_block();
 *        switch e B2 default raise(e)
 *     } else {
 *        B1;
 *        pop_try_block();
 *     }
 *   }
 *   { 
 *      e = pop_try_block();
 *      B3;
 *      if (e)
 *        raise(e);
 *   }
 */

/* Copy given float, so that it can be returned from function. */
double *make_float$(double v)
{
	double *ptr = xmalloc$(sizeof(*ptr));
	
	*ptr = v;
	
	return ptr;
}

/* Contruct new string. */
struct string$ *make_string$(ptroff size, char *content)
{
	struct string$ *r;

	r = xmalloc$(sizeof(*r) + size);
	r->len = size;
	r->data = (char*)r + sizeof(*r);
	memcpy(r->data, content, size);

	return r;
}

#include <unistd.h>

void print_string(struct string$ *s)
{
	write(1, s->data, s->len);
}
